unit FindTarg;
{
Highlight with blinking Drop control(s) at form(s) when Drag&Drop operation
took place. As a responce at message WM_LBUTTONDOWN, the application hook
scans all controls for assigned OnDragOver and send to them OnDragOver
message with Source object which lies at cursor and X,Y=-1,-1. Case
returned value of CanAccept=True the control is included in list and becomes
blinking to show anyone drag target.


Using: simple put the component at main form. OnDragOver procedures of controls
have to set appropriate value CanAccept with X,Y drag point= -1,-1.


FreeWare. No warranties. If one will distribute the source code, reference on
author (below) is required. If one will modify the source code, highlighting
(by any means) of modifications is required.

Author: Sergei Trepalin
E-mail: trep@trep.home.chg.ru}


interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ExtCtrls;

type
  TFindTarget = class(TComponent)
  private
    { Private declarations }
    FAppEvent:TMessageEvent;
    FDragControl:TControl;    {Stores current control which was started to drag}
    FDropList:TList;          {Possible trget for dragging control}
    FTimer:TTimer;
    FCanvas:TCanvas;
    FIsInverted:boolean;      {Has TRUE if targets are inverted}
    FTickCount:integer;
    FStartDrag,FEndDrag:TNotifyEvent; {Any can store own notifications in the methods}
    FControlStartDrag:TStartDragEvent;{Stores OnStartDrag for FDragControl}
    FControlEndDrag:TEndDragEvent;    {Stores OnEndDrag for FDragControl}
    procedure OnTimer(Sender:TObject);
    procedure PaintControls;
    function GetQuant:integer;              {Frequency of FDragControl highlighting change}
    procedure SetQuant(Value:integer);
    procedure UnhookDrag;                   {Release OnStartDrag,OnEndDrag for FDragControl}
    procedure HookDrag(Control:TControl);   {Hook OnStartDrag,OnEndDrag for FDragControl}
  protected
    { Protected declarations }
    procedure AppMessage(var Msg:TMsg; var Handled:boolean); virtual;
  public
    { Public declarations }
    constructor Create(AOwner:TComponent); override;
    destructor  Destroy; override;
    property TimeQuant:integer read GetQuant write SetQuant default 1000;
    procedure DragStart(Sender:TObject; var DragObject:TDragObject);
    procedure DragEnd(Sender,Target:TObject; X,Y:Integer);
  published
    { Published declarations }
    property StartDragNotify:TNotifyEvent read FStartDrag write FStartDrag;
    property EndDragNotify:TNotifyEvent read FEndDrag write FEndDrag;
  end;

procedure Register;
procedure InformStartDrag;
procedure InformEndDrag;

implementation

uses TypInfo;

type
  TMyControl=class(TControl)
  public
    property OnDragOver;
  end;

procedure InformStartDrag;
var
  Msg:TMsg;
  Handled:boolean;
begin
  if Assigned(Application.OnMessage) then begin
    Msg.Message:=WM_LBUTTONDOWN;
    Msg.hWnd:=0;
    Msg.wParam:=0;
    Msg.lParam:=0;
    Msg.Time:=0;
    GetCursorPos(Msg.Pt);
    Application.OnMessage(Msg,Handled);
  end;
end;

procedure InformEndDrag;
var
  Msg:TMsg;
  Handled:boolean;
begin
  if Assigned(Application.OnMessage) then begin
    Msg.Message:=WM_LBUTTONUP;
    Msg.hWnd:=0;
    Msg.wParam:=0;
    Msg.lParam:=0;
    Msg.Time:=0;
    GetCursorPos(Msg.Pt);
    Application.OnMessage(Msg,Handled);
  end;
end;

procedure CollectDragOver(ListTarget:TList);
var
  ListParent,ListChild:TList;
  I,J:integer;
  CM:TComponent;
begin
  if not Assigned(ListTarget) then Exit;
  ListTarget.Clear;
  ListParent:=nil;
  ListChild:=nil;
  try
    ListParent:=TList.Create;
    ListChild:=TList.Create;
    if Assigned(ListParent) and Assigned(ListChild) then begin
      for I:=0 to Application.ComponentCount-1 do ListParent.Add(Application.Components[I]);
      while ListParent.Count>0 do begin
        ListChild.Clear;
        for I:=0 to ListParent.Count-1 do if Assigned(ListParent[I]) then
        if TComponent(ListParent[I]).ComponentCount>0 then for J:=0 to TComponent(ListParent[I]).ComponentCount-1 do begin
          CM:=TComponent(ListParent[I]).Components[J];
          ListChild.Add(CM);
          if CM is TControl then if Assigned(TMyControl(CM).OnDragOver) then ListTarget.Add(CM);
        end;
        ListParent.Clear;
        if ListChild.Count>0 then for I:=0 to ListChild.Count-1 do ListParent.Add(ListChild[I]);
      end;
    end;
  finally
    if Assigned(ListParent) then ListParent.Free;
    if Assigned(ListChild) then ListChild.Free;
  end;
end;

{-----------------------------------------------------------------------------}
constructor TFindTarget.Create(AOwner:TComponent);
begin
  inherited Create(AOwner);
  if not (csDesigning in ComponentState) then begin
    FAppEvent:=Application.OnMessage;
    Application.OnMessage:=AppMessage;
  end;
  FDropList:=nil;
  FDragControl:=nil;
  FDropList:=TList.Create;
  FTimer:=TTimer.Create(Self);
  FTickCount:=0;
  if Assigned(FTimer) then begin
    FTimer.Interval:=1000;
    FTimer.OnTimer:=OnTimer;
    FTimer.Enabled:=False;
  end;
  FCanvas:=TCanvas.Create;
  FIsInverted:=False;
  FStartDrag:=nil;
  FEndDrag:=nil;
  FControlStartDrag:=nil;
  FControlEndDrag:=nil;
end;

destructor TFindTarget.Destroy;
begin
  UnhookDrag;
  if Assigned(FDropList) then FDropList.Free;
  if Assigned(FTimer) then FTimer.Free;
  if Assigned(FCanvas) then FCanvas.Free;
  if not (csDesigning in ComponentState) then Application.OnMessage:=FAppEvent;
  inherited Destroy;
end;

procedure TFindTarget.DragStart(Sender:TObject; var DragObject:TDragObject);
var
  I:integer;
  Test:boolean;
begin
  if Assigned(FControlStartDrag) then FControlStartDrag(Sender,DragObject);
  if Assigned(FDragControl) then begin
    CollectDragOver(FDropList);
    if FDropList.Count>0 then begin {Checking DragDrop}
      for I:=0 to FDropList.Count-1 do begin
        Test:=False;
        TMyControl(FDropList[I]).OnDragOver(TControl(FDropList[I]),FDragControl,-1,-1,dsDragMove,Test);
        if not Test then FDropList[I]:=nil;
      end;
      FDropList.Pack;
    end;
    if FDropList.Count>0 then begin
      if Assigned(FTimer) then FTimer.Enabled:=True;
      FIsInverted:=False;
      //PaintControls;
      if Assigned(FStartDrag) then FStartDrag(FDragControl);
    end else begin
      UnhookDrag;
      if not (csDesigning in ComponentState) then Application.Hint:='Can not find target';
      CancelDrag;
    end;
  end;
end;

procedure TFindTarget.DragEnd(Sender,Target:TObject; X,Y:Integer);
begin
  if Assigned(FControlEndDrag) then FControlEndDrag(Sender,Target,X,Y);
  UnhookDrag;
end;

procedure TFindTarget.UnhookDrag;
var
  PropInfo:PPropInfo;
  Method:TMethod;
  I:integer;
  Control:TControl;
begin
  FTimer.Enabled:=False;
  if Assigned(FDragControl) then begin
    if Assigned(FEndDrag) then FEndDrag(FDragControl);
    if FDropList.Count>0 then for I:=0 to FDropList.Count-1 do begin
      Control:=TControl(FDropList[I]);
      if Assigned(Control) then Control.Refresh;
    end;
    try
      PropInfo:=GetPropInfo(FDragControl.ClassInfo,'OnEndDrag');
      if Assigned(PropInfo) then begin
        Method:=TMethod(FControlEndDrag);
        SetMethodProp(FDragControl,PropInfo,Method);
      end;
      PropInfo:=GetPropInfo(FDragControl.ClassInfo,'OnStartDrag');
      if Assigned(PropInfo) then begin
        Method:=TMethod(FControlStartDrag);
        SetMethodProp(FDragControl,PropInfo,Method);
      end;
{
      TMyControl(FDragControl).OnStartDrag:=FControlStartDrag;
      TMyControl(FDragControl).OnEndDrag:=FControlEndDrag;
}
    except
    end;
    FDragControl:=nil;
  end;
  if Assigned(FDropList) then FDropList.Clear;
  FControlStartDrag:=nil;
  FControlEndDrag:=nil;
  FIsInverted:=False;
  FTickCount:=0;
 end;

procedure TFindTarget.HookDrag(Control:TControl);
var
  PropInfo:PPropInfo;
  DE:TEndDragEvent;
  DS:TStartDragEvent;
  Method:TMethod;
begin
  UnhookDrag;
  FDragControl:=Control;
  PropInfo:=GetPropInfo(Control.ClassInfo,'OnEndDrag');
  if Assigned(PropInfo) then begin
    FControlEndDrag:=TEndDragEvent(GetMethodProp(FDragControl,PropInfo));
    DE:=DragEnd;
    Method:=TMethod(DE);
    SetMethodProp(FDragControl,PropInfo,Method);
  end;
  PropInfo:=GetPropInfo(Control.ClassInfo,'OnStartDrag');
  if Assigned(PropInfo) then begin
    FControlStartDrag:=TStartDragEvent(GetMethodProp(FDragControl,PropInfo));
    DS:=DragStart;
    Method:=TMethod(DS);
    SetMethodProp(FDragControl,PropInfo,Method);
  end;
{
  FControlStartDrag:=TMyControl(Control).OnStartDrag;
  FControlEndDrag:=TMyControl(Control).OnEndDrag;
  TMyControl(Control).OnStartDrag:=DragStart;
  TMyControl(Control).OnEndDrag:=DragEnd;
}
end;


procedure TFindTarget.PaintControls;
var
  DC:HDC;
  I:integer;
  R:TRect;
  PW:TWinControl;
begin
  DC:=0;
  try
    if FDropList.Count>0 then for I:=0 to FDropList.Count-1 do begin
      DC:=0;
      if TObject(FDropList[I]) is TWinControl then begin
        {Windowed control - inverse all window}
        DC:=GetDC(TWinControl(FDropList[I]).Handle);
        FCanvas.Handle:=DC;
        if DC<>0 then begin
          GetClipBox(DC,R);
          InvertRect(DC,R);
          ReleaseDC(TWinControl(FDropList[I]).Handle,DC);
          DC:=0;
        end;
      end else if TObject(FDropList[I]) is TControl then begin
        PW:=TControl(FDropList[I]).Parent;
        if PW<>nil then begin
          DC:=GetDC(PW.Handle);
          FCanvas.Handle:=DC;
          if DC<>0 then begin
            R.Left:=TControl(FDropList[I]).Left;
            R.Right:=TControl(FDropList[I]).Left+TControl(FDropList[I]).Width;
            R.Top:=TControl(FDropList[I]).Top;
            R.Bottom:=TControl(FDropList[I]).Top+TControl(FDropList[I]).Height;
            InvertRect(DC,R);
            ReleaseDC(PW.Handle,DC);
            DC:=0;
          end;
        end;
      end;
    end;
  except
    FCanvas.Handle:=0;
    if DC<>0 then ReleaseDC(WindowFromDC(DC),DC);
  end;
  FIsInverted:=not FIsInverted;
end;

procedure TFindTarget.OnTimer(Sender:TObject);
begin
  FTickCount:=FTickCount+1;
  if FIsInverted or (FTickCount=1) then PaintControls else if (FTickCount mod 4)=0 then PaintControls;
end;

procedure TFindTarget.AppMessage(var Msg:TMsg; var Handled:boolean);
var
  CT:TControl;
  PropInfo:PPropInfo;
  DM:TDragMode;
begin
  if (Msg.Message=WM_LBUTTONDOWN) and Assigned(FDropList) then begin
    CT:=FindDragTarget(Msg.PT,False);
    if Assigned(CT) then begin
      PropInfo:=GetPropInfo(CT.ClassInfo,'DragMode');
      if Assigned(PropInfo) then begin
        DM:=TDragMode(GetOrdProp(CT,PropInfo));
        if (DM=dmAutomatic) or ((Msg.hWnd=0) and (Msg.wParam=0) and (Msg.lParam=0)
        and (Msg.Time=0)) then HookDrag(CT);
      end;
    end;
  end else if (Msg.Message=WM_LBUTTONUP) then UnhookDrag
  else if (Msg.Message=WM_MOUSEMOVE) and Assigned(FDragControl) and (not FDragControl.Dragging) then UnhookDrag;
  if Assigned(FAppEvent) then FAppEvent(Msg,Handled);
end;

function TFindTarget.GetQuant:integer;
begin
  Result:=0;
  if Assigned(FTimer) then Result:=FTimer.Interval;
end;

procedure TFindTarget.SetQuant(Value:integer);
begin
  if Assigned(FTimer) then FTimer.Interval:=Value;
end;


procedure Register;
begin
  RegisterComponents('Interface', [TFindTarget]);
end;

end.
